/*
*  C Implementation: "Bootloader" for ATTiny2313
*
* Description:
*
*  Bootloader somewhat compatible with "butterfly" "avr109" and "avr910" modes of avrdude
*
*  The attiny2313 has no direct bootloader support, so the application using it has to invoke the bootloader manually.
*  if the application already uses the serial port, listen for 0x1B 'S', and invoke "bootloader(1)"
*  otherwise, remove the "NO_SYNC" define below, and invoke "bootloader(0)"
*  avr910 doesn't use that escape code, here always bootloader(1) is needed.
*
*  you can create a library from this code and link your app against it, or directly flash a .hex from it.
*
*  applications using this bootloader need to link against the lib anyways, so it can call the bootloader at all.
*
*  the write-flash code takes care that the bootloader itself can't be overwritten.
*
*  link the resulting binary with LDFLAGS += -Wl,--section-start=.bootloader=0x500
*
*  This code is far from beeing optimized, it uses up almost half of the flash. Improved versions welcome.
*
*  This code uses the WDT to reset after flashing, the main application code is responsible for stopping the WDT again
*  if it doesn't use it. (WDT keeps running after reset)
*
* Author: Ernst Bachmann <e.bachmann@netandweb.de>, (C) 2006
*
* Copyright: See COPYING file that comes with this distribution
*
*/

#include <avr/io.h>
#include <avr/interrupt.h>

// Buggy iotn2313.h
#define SPMEN SELFPRGE
#define BLBSET RFLB
#include <avr/boot.h>

//#include <avr/sleep.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include <stdint.h>

//#define SERIAL_WAIT (30000)
#define NO_SYNC

#define BAUD 115200
#define UART_BAUD_CALC(BAUD_RATE) ((uint16_t)(((F_OSC)/((BAUD_RATE*8.0))+0.5)-1))

static void boot_sendchar(char data) BOOTLOADER_SECTION;

// //static void boot_sendstring_P(const prog_char * p) BOOTLOADER_SECTION;

static char boot_readchar(void) BOOTLOADER_SECTION;


void bootloader(uint8_t sync) BOOTLOADER_SECTION __attribute__((noreturn));
void bootloader(uint8_t sync) {
  uint8_t i=0;
  char ch;

  union {
    uint8_t * ptr;
    uint16_t word;
    struct {
      uint8_t lo;
      uint8_t hi;
    };
  } address;

  uint16_t size;


  cli(); // no irqs.

  // set Baud
  UBRRH=UART_BAUD_CALC(BAUD)>>8;
  UBRRL=UART_BAUD_CALC(BAUD)&0xFF;

  // at Double Speed
  UCSRA = _BV(U2X);

  // with 8N1 Frames
  UCSRC = _BV(UCSZ0)|_BV(UCSZ1);

  // Transmitter + Receiver enable
  UCSRB = _BV(TXEN) | _BV(RXEN);

#ifndef NO_SYNC
  // Try to sync with loader
   while (!sync) {
     ch = boot_readchar();
     i++;
     if(ch == 0x1B) { // ESC
       do {
         ch=boot_readchar();
         i++;
      } while (ch != 'S');
      sync=1;
     } else {
        boot_sendchar('?');
     }
  }
#endif
  ch='S';
  while (1) {
      if (ch == 'a') { // Auto-Increment?

        boot_sendchar('Y'); // Yes!

      } else if (ch == 'p') { // get Programmer Type

        boot_sendchar('S'); // Serial

      } else if (ch == 'S') { // get Software

        boot_sendchar('E');
        boot_sendchar('G');
        boot_sendchar('B');
        boot_sendchar('B');
        boot_sendchar('O');
        boot_sendchar('O');
        boot_sendchar('T');

      } else if (ch == 'V' || ch =='v') { // get Version

        boot_sendchar('0');
        boot_sendchar('a');

      } else if (ch == 'A') { // Set Write Address

        address.hi=boot_readchar();
        address.lo=boot_readchar();
        boot_sendchar('\r'); // ACK

      } else if (ch == 's') {// Return Signature Byte

        boot_sendchar(0x1E);
        boot_sendchar(0x91);
        boot_sendchar(0x0A);

     } else if (ch == 'b') {// Return BufferSize

        boot_sendchar('Y');
        boot_sendchar(0);
        boot_sendchar(SPM_PAGESIZE);

      } else if (ch == 't') { // get supported Device Types

        boot_sendchar(0x20); // AT90S2313
        boot_sendchar(0x23); // ATTiny 2313
        boot_sendchar(0);

      } else if (ch == 'T') { // set Device Types: ignore but ACK

	boot_readchar();
        boot_sendchar('\r');

      } else if (ch == 'F') { // Read L-Fuses

	boot_sendchar(boot_lock_fuse_bits_get(GET_LOW_FUSE_BITS));

      } else if (ch == 'r') { // Read Lock Bits

	boot_sendchar(boot_lock_fuse_bits_get(GET_LOCK_BITS));

      } else if (ch == 'N') { // Read H-Fuses

	boot_sendchar(boot_lock_fuse_bits_get(GET_HIGH_FUSE_BITS));

      } else if (ch == 'Q') { // Read H-Fuses

	boot_sendchar(boot_lock_fuse_bits_get(GET_EXTENDED_FUSE_BITS));

     } else if (ch == 'g') { // Read memory block

        size = boot_readchar() << 8;	// high byte of buffersize
	size |= boot_readchar();	// low byte
        ch = boot_readchar();   // memory type

        if (ch == 'E') { // EEPROM

	  eeprom_busy_wait();
	  while (size--) {

//            boot_sendchar(eeprom_read_byte(address.ptr++));

	    EEAR = address.lo++;   // Setup EEPROM address (only low-byte)
            EECR |= _BV(EERE);
	    boot_sendchar(EEDR);
          }

        } else { // FLASH

          address.word <<= 1; // 16bit (instruction)-addr => Byte addr
/*
	  // AVRdude expects it in the "wrong" order
	  while (size) {
            boot_sendchar(pgm_read_byte(address.ptr+1));
            boot_sendchar(pgm_read_byte(address.ptr));
	    address.ptr += 2;
	    size -= 2;
          }
*/
	  while (size--) {
	    boot_sendchar(pgm_read_byte(address.ptr++));
          }
          // restore "instruction" address for auto-increment mode
	  address.word >>= 1;
        }

      } else if (ch == 'e') { // Chip Erase

        for(address.word=0;address.word < APP_END;address.word += SPM_PAGESIZE) {
          boot_page_erase(address.ptr);
          boot_spm_busy_wait();
        }

        boot_sendchar('\r'); // ACK

      } else if (ch == 'B') { // Programm with buffer

        size = boot_readchar() << 8;	// high byte of buffersize
	size |= boot_readchar();	// low byte
        ch = boot_readchar();   // memory type

        if (ch == 'E')	{ // EEPROM

          while (size--) {
//            boot_eeprom_write_byte(address.ptr++,boot_readchar());
              EEAR=address.lo++;
              EEDR=boot_readchar();
              EECR |= _BV(EEMPE); // Enable programming
              EECR |= _BV(EEPE); // Start write
              eeprom_busy_wait();  // Wait till finished
          }
          boot_sendchar('\r'); // ACK

        } else if ((ch == 'F') && !(address.word & (SPM_PAGESIZE-1))) { // FLASH

          if (address.word < APP_END) {
            while (size) {
                i=boot_readchar(); // High byte of instruction
                boot_page_fill(address.ptr,i|(boot_readchar()<<8));
                address.ptr+=2;
                size-=2;
            }
            // Fill remaining bank with 0xFF
            while (address.word & (SPM_PAGESIZE-1)) {
                boot_page_fill(address.ptr, 0xFFFF);
                address.ptr+=2;
            }
            boot_page_write(address.ptr-SPM_PAGESIZE);
            boot_spm_busy_wait();
          } else {
            // Fake programming for the bootloader section
            while (size--) {
              boot_readchar();
            }
            address.word+=SPM_PAGESIZE;

          }
          boot_sendchar('\r'); // ACK

        } else {
          boot_sendchar(0); // Unsupported memory type or wrong params
        }

      } else if (ch == 'E') { // rEboot

        boot_sendchar('\r'); // ACK
        // Cause Reset:
        wdt_enable(WDTO_30MS);
        while(1); // Wait for doomsday

      } else if ((ch == 'P') || (ch == 'L')) { // Enter/Leave Programming mode, NOP but ACK

        boot_sendchar('\r');

/*
      } else if ((ch == 'x') || (ch == 'y')) { // clear+set LEDs from byte. Ignore, but read byte and ack

        PIND|=_BV(PD6);
        boot_readchar();
        boot_sendchar('\r');
*/
      } else if ((ch == 0x1B) || (ch == '\n')) { // ESC,\n: nop
      } else {

        boot_sendchar('?');

      }
      ch = boot_readchar();

  }
}



static void boot_sendchar(char data) {
//    uint16_t i = 0;

//    while ( !(UCSRA&_BV(UDRE)) && (i<SERIAL_WAIT) ) i++;
    loop_until_bit_is_set(UCSRA,UDRE);
    UDR = data;

}

/*
static void boot_sendstring_P(const char * p) {
  char c;

  c=pgm_read_byte(p++);
  while (c) {
    boot_sendchar(c);
    c=pgm_read_byte(p++);
  }
}
*/

static char boot_readchar(void) {
    //uint16_t i = 0;
//    while(!(UCSRA & _BV(RXC)) && (i< SERIAL_WAIT)) i++;
//    if (i < SERIAL_WAIT)
      loop_until_bit_is_set(UCSRA,RXC);
      return UDR;
//    return 0x00;
}


/* dummy main. will be overwritten by real main */

int main(void) __attribute__ ((weak));

int main(void) {
  bootloader(1);
}



